"""Arraymapper module

.. rubric:: Objects

.. autosummary::
   :toctree:

   x

"""

from functools import wraps
from typing import Callable

from gammy.utils import compose


class ArrayMapper:
    """Callable input mapping object that obey arithmetic operations

    When working with basis function regression, when building the
    design matrix, one needs the information 'how eact term maps input array'.
    This class helps writing human-readable model definitions that explicitly
    show how inputs are mapped. For example,

    .. code-block:: python

       formula = a(x[:, 0]) * x[:, 1] + x[:, 2] ** 2

    is a valid definition as long as ``a`` is an instance of
    :class:`gammy.Formula` and ``x`` is an instance of :class:`ArrayMapper`.

    TODO: Are some natural methods are still missing?

    """

    def __init__(self, function=lambda t: t):

        self.function = function
        """Wrapped function"""

        return

    def __call__(self, *args, **kwargs):
        """Call the wrapped function

        """
        return self.function.__call__(*args, **kwargs)

    def __getitem__(self, key: int) -> "ArrayMapper":
        """Access index as in a NumPy Array

        """
        return ArrayMapper(
            lambda t: t.__getitem__(key)
        )

    def __add__(self, other) -> "ArrayMapper":
        """Addition of arraymappers

        Parameters
        ----------
        other : ArrayMapper

        """
        return ArrayMapper(
            lambda t: self.function(t).__add__(other.function(t))
        )

    def __sub__(self, other) -> "ArrayMapper":
        """Subtraction of arraymappers

        Parameters
        ----------
        other : ArrayMapper

        """
        return ArrayMapper(
            lambda t: self.function(t).__sub__(other.function(t))
        )

    def __mul__(self, other) -> "ArrayMapper":
        """Multiplication of arraymappers

        Parameters
        ----------
        other : ArrayMapper

        """
        return ArrayMapper(
            lambda t: self.function(t).__mul__(other.function(t))
        )

    def __truediv__(self, other) -> "ArrayMapper":
        """Division of arraymappers

        Parameters
        ----------
        other : ArrayMapper

        """
        return ArrayMapper(
            lambda t: self.function(t).__truediv__(other.function(t))
        )

    def __pow__(self, n: float) -> "ArrayMapper":
        """Raise an arraymapper to a power

        """
        return ArrayMapper(
            lambda t: self.function(t).__pow__(n)
        )

    def __neg__(self) -> "ArrayMapper":
        """Negation operation

        """
        return ArrayMapper(
            lambda t: self.function(t).__neg__()
        )

    def ravel(self) -> "ArrayMapper":
        """Imitates the behavior of NumPy ravel method

        """
        return ArrayMapper(
            lambda t: self.function(t).ravel()
        )

    def reshape(self, *args, **kwargs) -> "ArrayMapper":
        """Numpy reshape

        """
        return ArrayMapper(
            lambda t: self.function(t).reshape(*args, **kwargs)
        )


def lift(f):
    """Lift a function

    lift :: a -> b -> ArrayMapper(a) -> ArrayMapper(b) -> ArrayMapper(b)

    """

    @wraps(f)
    def lifted(mapper):
        return ArrayMapper(compose(f, mapper.function))

    return lifted


x = ArrayMapper()
x.__doc__ = (
    """ArrayMapper instantiated at load-time for convenient importing to applications

    Intended usage is to import ``x`` standalone:

    .. code-block:: python

        import gammy
        from gammy.arraymapper import x

        # Define formula
        formula = gammy.Scalar() * x

    """
)



"""Plotting tools for Gammy models and formulae

"""

import logging
from typing import Callable, List, Optional, Union

import bayespy as bp
try:
    from mpl_toolkits.mplot3d import Axes3D
except ImportError:
    logging.info(
        "Problem with importing Axes3D from mpl_toolkits.mplot3d. Skipping."
    )
import matplotlib.pyplot as plt
from matplotlib.colors import SymLogNorm
import numpy as np

import gammy
from gammy import utils
from gammy.utils import pipe


def validation_plot(
        model,
        input_data,
        y,
        grid_limits,
        input_maps,
        index=None,
        xlabels=None,
        titles=None,
        gridsize=20,
        color="r",
        **kwargs
):
    """Validation plot for a GAM object

    Contains:

        - Series plot with predicted vs. observed
        - Partial residual plots

    Parameters
    ----------
    model : gammy.bayespy.GAM | gammy.numpy.GAM
        Visualized model
    input_data : np.ndarray
        Input data
    y : np.ndarray
        Observations
    grid_limits : List
        Grid limits, either `[a, b]` or `[[a_1, b_1], ..., [a_N, b_N]]`
    input_maps : List[Callable]
        List of input maps to be used for each pair of grid limits
    index : np.ndarray
        Optional x-axis for the series plot
    xlabels : List[str]
        Optional x-labels for the partial residual plots
    gridsize : int
        Number of points in the input dimensions discretizations
    color : str
        Color of scatter points

    """

    N = len(model.formula)
    N_rows = 2 + (N + 1) // 2
    fig = plt.figure(figsize=(8, 2 * N_rows))
    gs = fig.add_gridspec(2 + (N + 1) // 2, 2)
    xlabels = xlabels or [None] * len(model.formula)
    titles = titles or [None] * len(model.formula)
    index = np.arange(len(input_data)) if index is None else index

    assert (
        len(grid_limits) == 2 if len(input_data.shape) == 1 else
        (
            len(grid_limits) == input_data.shape[1] and
            all([len(xs) == 2 for xs in grid_limits])
        )
    ), (
        "Given grid limits do not match with the shape of input data."
    )
    assert len(model.formula.terms) == len(input_maps), (
        "Must give exactly one input per model term."
    )
    assert len(model.formula.terms) == len(titles), (
        "Must give exactly one title per model term."
    )

    # Data and predictions
    grid = np.array(
        utils.listmap(lambda x: np.linspace(x[0], x[1], gridsize))(grid_limits)
    ).T if len(input_data.shape) == 2 else np.linspace(
        grid_limits[0], grid_limits[1], gridsize
    )
    marginals = model.predict_variance_marginals(grid)
    residuals = model.marginal_residuals(input_data, y)

    # Time series plot
    ax = fig.add_subplot(gs[0, :])
    (mu, sigma_theta) = model.predict_variance_theta(input_data)
    lower = mu - 2 * np.sqrt(sigma_theta + model.inv_mean_tau)
    upper = mu + 2 * np.sqrt(sigma_theta + model.inv_mean_tau)
    ax.plot(index, y, linewidth=0, marker="o", alpha=0.3, color=color)
    ax.plot(index, mu, color="k")
    ax.fill_between(index, lower, upper, color="k", alpha=0.3)
    ax.grid(True)

    # XY-plot
    ax = fig.add_subplot(gs[1, :])
    ax.plot(mu, y, alpha=0.3, marker="o", lw=0, color=color)
    ax.plot([mu.min(), mu.max()], [mu.min(), mu.max()], c="k", label="x=y")
    ax.legend(loc="best")
    ax.grid(True)
    ax.set_xlabel("Predictions")
    ax.set_ylabel("Observations")

    # Partial residual plots
    for i, ((mu, sigma), res, input_map, xlabel, title) in enumerate(
        zip(marginals, residuals, input_maps, xlabels, titles)
    ):
        x = input_map(grid)
        if len(x.shape) == 1 or x.shape[1] == 1:
            ax = fig.add_subplot(gs[2 + i // 2, i % 2])
            (lower, upper) = (
                mu - 2 * np.sqrt(sigma),
                mu + 2 * np.sqrt(sigma)
            )
            ax.scatter(input_map(input_data), res, color=color, **kwargs)
            ax.plot(x, mu, c='k', lw=2)
            ax.fill_between(x, lower, upper, alpha=0.3, color="k")
            ax.set_xlabel(xlabel)
        elif x.shape[1] == 2:
            ax = fig.add_subplot(gs[2 + i // 2, i % 2], projection="3d")
            u, v = np.meshgrid(x[:, 0], x[:, 1])
            w = np.hstack((
                u.reshape(-1, 1), v.reshape(-1, 1)
            ))
            # Override mu and sigma on purpose!
            (mu, sigma) = model.predict_variance_marginal(w, i)
            mu_mesh = mu.reshape(u.shape)
            ax.plot_surface(u, v, mu_mesh)
        else:
            raise NotImplementedError("High-dimensional plots not supported.")
        ax.set_title(title)
        ax.grid(True)

    fig.tight_layout()
    return fig


def gaussian1d_density_plot(model: gammy.bayespy.GAM):
    """Plot 1-D density for each parameter

    """
    N = len(model.formula)
    N_rows = 2 + (N + 1) // 2
    fig = plt.figure(figsize=(8, 2 * N_rows))
    gs = fig.add_gridspec(N + 1, 1)

    # Plot inverse gamma
    ax = fig.add_subplot(gs[0])
    (b, a) = (-model.tau.phi[0], model.tau.phi[1])
    mu = a / b
    grid = np.arange(0.5 * mu, 1.5 * mu, mu / 300)
    ax.plot(grid, model.tau.pdf(grid))
    ax.set_title(r"$\tau$ = noise inverse variance")
    ax.grid(True)

    # Plot marginal thetas
    for i, theta in enumerate(model.theta_marginals):
        ax = fig.add_subplot(gs[i + 1])
        mus = theta.get_moments()[0]
        mus = np.array([mus]) if mus.shape == () else mus
        cov = utils.solve_covariance(theta.get_moments())
        stds = pipe(
            np.array([cov]) if cov.shape == ()
            else np.diag(cov),
            np.sqrt
        )
        left = (mus - 4 * stds).min()
        right = (mus + 4 * stds).max()
        grid = np.arange(left, right, (right - left) / 300)
        for (mu, std) in zip(mus, stds):
            node = bp.nodes.GaussianARD(mu, 1 / std ** 2)
            ax.plot(grid, node.pdf(grid))
        ax.set_title(r"$\theta_{0}$".format(i))
        ax.grid(True)

    fig.tight_layout()
    return fig


def gaussian2d_density_plot(model: gammy.bayespy.GAM, i, j):
    """Plot 2-D joint distribution of indices i and j

    """
    raise NotImplementedError


def covariance_plot(model, ax=None, linthresh=0.1, **kwargs):
    """Covariance matrix

    """

    ax = plt.figure().gca() if ax is None else ax

    C = model.covariance_theta
    im = ax.imshow(
        C,
        norm=SymLogNorm(
            vmin=np.min(C),
            vmax=np.max(C),
            linthresh=linthresh
        ),
        **kwargs
    )

    return (ax, im)


def basis_plot(
        formula: gammy.formulae.Formula,
        grid_limits,
        input_maps,
        gridsize=20
):
    """Plot all basis functions

    """
    # Figure definition
    N = len(formula)
    fig = plt.figure(figsize=(8, max(4 * N // 2, 8)))
    gs = fig.add_gridspec(N, 1)

    # Data and predictions
    grid = (
        pipe(
            grid_limits,
            utils.listmap(lambda x: np.linspace(x[0], x[1], gridsize)),
            lambda x: np.array(x).T
        )
    )

    # Plot stuff
    for i, (basis, input_map) in enumerate(zip(formula.terms, input_maps)):
        ax = fig.add_subplot(gs[i])
        x = input_map(grid)
        for f in basis:
            ax.plot(x, f(grid))

    return fig



"""Formula type definition and constructors

"""


from typing import Callable, List, Tuple, Iterable

import numpy as np
import scipy as sp
from scipy import interpolate

from gammy import utils
from gammy.arraymapper import ArrayMapper
from gammy.utils import listmap, rlift


class Formula:
    """Basis manipulation and design matrix creator

    Parameters
    ----------
    terms : List[Callable]
        Each element is a list of basis functions and corresponds to a term
        in the additive model formula
    prior : Tuple[np.ndarray]
        Mean and precision matrix of the Gaussian prior distribution

    """

    def __init__(self, terms, prior):

        self.terms = terms
        """List of basis functions """

        self.prior = prior
        """Prior mean and precision"""

        return


    def __add__(self, other) -> "Formula":
        """Addition of formulae

        Parameters
        ----------
        other : Formula

        """
        return Formula(
            terms=self.terms + other.terms,
            prior=utils.concat_gaussians([self.prior, other.prior])
        )

    def __mul__(self, input_map) -> "Formula":
        """Multiplication

        Parameters
        ----------
        input_map : ArrayMapper

        """
        return Formula(
            terms=[
                listmap(
                    lambda f: lambda t: f(t) * input_map(t)
                )(basis) for basis in self.terms
            ],
            prior=self.prior
        )

    def __len__(self) -> int:
        """Number of terms in the formula

        """
        return len(self.terms)

    def __call__(self, *input_maps) -> "Formula":
        """Make the object callable

        Parameters
        ----------
        input_maps : Iterable[ArrayMapper]

        """
        # TODO: Transform basis
        return Formula(
            terms=[
                listmap(rlift(input_map))(basis)
                for (basis, input_map) in zip(self.terms, input_maps)
            ],
            prior=self.prior
        )

    def design_matrix(self, input_data, i: int=None):
        # If one term is asked for, give it. Otherwise use all terms.
        fs = sum(self.terms, []) if i is None else self.terms[i]
        return np.hstack([f(input_data).reshape(-1, 1) for f in fs])


#
# Operations between formulae
#


def Flatten(formula, prior=None) -> Formula:
    """Flatten the terms of a given formula

    Optionally override prior

    Parameters
    ----------
    formula : Formula
        Flattened formula with a nested list of terms
    prior : Tuple[np.ndarray]
        Prior of the final formula

    In terms of terms: ``[[f1, f2], [g1, g2, g3]] => [[f1, f2, g1, g2, g3]]``

    """
    return Formula(
        terms=[sum(formula.terms, [])],
        prior=formula.prior if prior is None else prior
    )


def Sum(formulae, prior=None) -> Formula:
    """Sum (concatenate) many formulae

    Parameters
    ----------
    formulae : List[Formula]
        Formulas to concatenate
    prior : Tuple[np.ndarray]
        Prior mean and covariance for concatenated formula

    Theoretical example:

    .. code-block:: text

        ([[f1, f2], [g1, g2]], [[h1]]) => [[f1, f2], [g1, g2], [h1]]

    NOTE: :class:`Sum` and :class:`Flatten` are different!

    """
    priors = [formula.prior for formula in formulae]
    return Formula(
        terms=sum([formula.terms for formula in formulae], []),
        prior=utils.concat_gaussians(priors) if prior is None else prior
    )


def Kron(a, b) -> Formula:
    """Tensor product of two Formula terms

    Parameters
    ----------
    a : Formula
        Left input
    b : Formula
        Right input

    Non-commutative!

    Let ``u, v`` be eigenvectors of matrices ``A, B``, respectively. Then
    ``u ⊗ v`` is an eigenvector of ``A ⊗ B`` and ``λμ`` is the corresponding
    eigenvalue.

    """
    # NOTE: This is somewhat experimental. The terms must correspond to
    #       "zero-mean" r.v.. Then Kronecker product of covariances
    #       corresponds to the product r.v. of independent r.v.'s.
    #       Check the formula of variance of product of independent r.v.'s.

    # TODO / FIXME: Don't flatten a and b.
    gen = (
        # Careful! Must be same order as in a Kronecker product.
        (f, g) for f in sum(a.terms, []) for g in sum(b.terms, [])
    )

    # Outer product of terms
    basis = listmap(
        lambda funcs: lambda t: funcs[0](t) * funcs[1](t)
    )(gen)

    # Kronecker product of prior means and covariances
    return Formula(
        terms=[basis],
        prior=(
            np.kron(a.prior[0], b.prior[0]),
            # Although we kron-multiply precision matrices here (inverse
            # of covariance), the order of inputs doesn't flip because
            # (A ⊗ B) ^ -1 = (A ^ -1) ⊗ (B ^ -1)
            np.kron(a.prior[1], b.prior[1])
        )
    )


#
# Custom formulae collection
#


def create_from_kernel1d(kernel: Callable) -> Callable:
    """Create formula from bivariate GP kernel function

    """

    def _Formula(
            grid: np.ndarray,
            prior: Tuple[np.ndarray]=None,
            mu_basis: List[Callable]=None,
            mu_hyper: Tuple[np.ndarray]=None,
            energy: float=.99,
            **kernel_kwargs
    ) -> Formula:

        mu_basis = [] if mu_basis is None else mu_basis
        basis = utils.interp_arrays1d(
            utils.decompose_covariance(
                kernel(
                    x1=grid.reshape(-1, 1),
                    x2=grid.reshape(-1, 1),
                    **kernel_kwargs
                ),
                energy=energy
            ),
            grid=grid,
            fill_value="extrapolate"
        )

        # Default prior is white noise for the problem with
        # the constructed basis functions
        prior = (
            (np.zeros(len(basis)), np.identity(len(basis)))
            if prior is None else prior
        )

        return Formula(
            terms=[mu_basis + basis],
            prior=prior if mu_hyper is None else utils.concat_gaussians(
                [mu_hyper, prior]
            )
        )

    return _Formula


def ExpSquared1d(
        grid,
        corrlen,
        sigma,
        prior=None,
        mu_basis=None,
        mu_hyper=None,
        energy=0.99
) -> Formula:
    """Squared exponential kernel formula

    Example
    -------

    .. code-block:: python

        formula = ExpSquared1d(
            grid=np.arange(-25, 35, 1.0),
            l=5.0,
            sigma=1.0,
            mu_basis=[lambda t: t],
            mu_hyper=(0, 1e-6)
        )

    """
    kernel_kwargs = {
        "corrlen": corrlen,
        "sigma": sigma
    }
    _Formula = create_from_kernel1d(utils.exp_squared)
    return _Formula(
        grid=grid,
        prior=prior,
        mu_basis=mu_basis,
        mu_hyper=mu_hyper,
        energy=energy,
        **kernel_kwargs
    )


def ExpSineSquared1d(
        grid,
        corrlen,
        sigma,
        period,
        prior=None,
        mu_basis=None,
        mu_hyper=None,
        energy=0.99
) -> Formula:
    """Squared sine exponential kernel formula for periodic terms

    Parameters
    ----------
    grid : np.ndarray
        Discretization grid
    corrlen : float
        Correlation length
    sigma : float
        Variance
    period : float
        Period
    prior : Tuple[np.ndarray]
        Prior mean and precision matrix
    mu_basis : List[Callable]
        Basis for estimating the mean hyperparameter
    mu_hyper : Tuple[np.ndarray]
        Hyperprior mean and precision matrix
    energy : float
        Eigenvalue-weighted proportion of eigenvectors to consider in
        truncation

    """
    kernel_kwargs = {
        "corrlen": corrlen,
        "sigma": sigma,
        "period": period
    }
    _Formula = create_from_kernel1d(utils.exp_sine_squared)
    return _Formula(
        grid=grid,
        prior=prior,
        mu_basis=mu_basis,
        mu_hyper=mu_hyper,
        energy=energy,
        **kernel_kwargs
    )


def RationalQuadratic1d(
        grid,
        corrlen,
        sigma,
        alpha,
        prior=None,
        mu_basis=None,
        mu_hyper=None,
        energy=0.99
) -> Formula:
    """Rational quadratic kernel formula

    """
    kernel_kwargs = {
        "corrlen": corrlen,
        "sigma": sigma,
        "alpha": alpha
    }
    _Formula = create_from_kernel1d(utils.rational_quadratic)
    return _Formula(
        grid=grid,
        prior=prior,
        mu_basis=mu_basis,
        mu_hyper=mu_hyper,
        energy=energy,
        **kernel_kwargs
    )


def WhiteNoise1d(
        grid,
        sigma,
        prior=None,
        mu_basis=None,
        mu_hyper=None,
        energy=1.0,
) -> Formula:
    """White noise kernel formula

    """
    kernel_kwargs = {
        "n_dims": len(grid),
        "sigma": sigma
    }
    _Formula = create_from_kernel1d(utils.white_noise)
    return _Formula(
        grid=grid,
        prior=prior,
        mu_basis=mu_basis,
        mu_hyper=mu_hyper,
        energy=energy,
        **kernel_kwargs
    )


def OrnsteinUhlenbeck1d(
        grid,
        corrlen,
        sigma,
        prior=None,
        mu_basis=None,
        mu_hyper=None,
        energy=0.99
) -> Formula:
    """Ornstein-Uhlenbeck kernel formula

    """
    kernel_kwargs = {
        "corrlen": corrlen,
        "sigma": sigma,
    }
    _Formula = create_from_kernel1d(utils.ornstein_uhlenbeck)
    return _Formula(
        grid=grid,
        prior=prior,
        mu_basis=mu_basis,
        mu_hyper=mu_hyper,
        energy=energy,
        **kernel_kwargs
    )


def Scalar(prior: Tuple[np.ndarray]=(0, 1e-6)) -> Formula:
    """Scalar formula

    """
    basis = [lambda t: np.ones(len(t))]
    return Formula(terms=[basis], prior=prior)


def Function(function: Callable, prior: Tuple[np.ndarray]) -> Formula:
    """Construct a formula from a single function

    """
    basis = [function]
    return Formula(terms=[basis], prior=prior)


def Polynomial(degree, prior=None):

    def monomial(p):
        return lambda t: t ** p

    basis = [monomial(n) for n in range(degree + 1)]
    prior = (np.zeros(degree + 1), 1e-6 * np.identity(degree + 1))
    return Formula(terms=[basis], prior=prior)


def ReLU(grid: np.ndarray, prior: Tuple[np.ndarray]=None) -> Formula:
    """Rectified linear unit shaped basis

    """
    relus = listmap(lambda c: lambda t: (t > c) * (c - t))(grid[1:-1])
    prior = (
        (np.zeros(len(grid) - 2), np.identity(len(grid) - 2))
        if not prior else prior
    )
    return Formula(terms=[relus], prior=prior)


def FlippedReLU(grid: np.ndarray, prior: Tuple[np.ndarray]=None) -> Formula:
    """Mirrored ReLU basis

    """
    relus = listmap(lambda c: lambda t: (t < c) * (c - t))(grid[1:-1])
    prior = (
        (np.zeros(len(grid) - 2), np.identity(len(grid) - 2))
        if not prior else prior
    )
    return Formula(terms=[relus], prior=prior)


def TanH() -> Formula:
    raise NotImplementedError


def Gaussian1d() -> Formula:
    raise NotImplementedError


def BSpline1d(
        grid,
        order=3,
        extrapolate=True,
        prior=None,
        mu_basis=None,
        mu_hyper=None
) -> Formula:
    """B-spline basis on a fixed one-dimensional grid

    Parameters
    ----------
    grid : np.ndarray
        Discretization grid
    order : int
        Order of the spline function. Polynomial degree is ``order - 1``
    extrapolate : bool
        Extrapolate outside of the grid using basis functions "touching" the
        endpoints
    prior : Tuple[np.ndarray]
        Prior mean and precision matrix
    mu_basis : List[Callable]
        Basis for estimating the mean hyperparameter
    mu_hyper : Tuple[np.ndarray]
        Hyperprior mean and precision matrix

    Number of spline basis functions is always ``N = len(grid) + order - 2``

    TODO: Verify that this doesn't break when scaling the grid
          (extrapolation + damping)

    """

    mu_basis = [] if mu_basis is None else mu_basis
    grid_ext = utils.extend_spline_grid(grid, order)

    def build_basis_element(spline_arg):

        (knots, extrapolate, loc) = spline_arg

        def right_damp(t):
            return t > knots[-1]

        def left_damp(t):
            return knots[0] > t

        def element(t):
            sp_element = interpolate.BSpline.basis_element(
                knots,
                extrapolate if loc in (-1, 1) else False
            )
            return sp_element(t) if loc == 0 else (
                sp_element(t) * right_damp(t) if loc == -1 else
                sp_element(t) * left_damp(t)
            )

        return utils.compose2(np.nan_to_num, element)

    basis = listmap(build_basis_element)(
        utils.gen_spline_args_from_grid_ext(grid_ext, order, extrapolate)
    )

    # Default prior is white noise
    prior = (
        (np.zeros(len(basis)), 1e-6 * np.identity(len(basis)))
        if prior is None else prior
    )
    return Formula(
        terms=[mu_basis + basis],
        prior=prior if mu_hyper is None else utils.concat_gaussians(
            [mu_hyper, prior]
        )
    )



"""BayesPy engine

"""


import json
from typing import Callable, List, Tuple

import bayespy as bp
import h5py
import numpy as np

import gammy
from gammy import utils
from gammy.formulae import Formula


def create_gaussian_theta(formula: Formula):
    return bp.nodes.Gaussian(*formula.prior)


class GAM:
    """Generalized additive model with BayesPy backend

    Parameters
    ----------
    formula : gammy.formulae.Formula
        Formula object containing the terms and prior
    theta : bp.nodes.Gaussian
        Model parameters vector
    tau : bp.nodes.Gamma
        Observation noise precision (inverse variance)

    Currently tau is fixed to Gamma distribution, i.e., it is not
    possible to manually define the noise level. Note though that one can
    set tight values for `α, β` in `Gamma(α, β)`, recalling that `mean = α / β`
    and `variance = α / β ** 2`. The upside is that by estimating the noise
    level, one gets a nice prediction uncertainty estimate.

    Currently Gaussian requirement deeply built in. Tau being Gamma
    implies, by conjugacy, that theta must be Gaussian.

    Does not support scalar valued Gaussian r.v.. Could be implemented
    using GaussianARD but this would require a lot of refactoring for such a
    small feature -- after all one can define an auxiliary bias term with a
    very tight prior.

    TODO: Statistics for basis function evaluations at grid points.

    FIXME: BayesPy fit fails with the following example:

        .. code-block:: python

           gammy.bayespy.GAM(
               gammy.Scalar()
           ).fit(np.array([0]), np.array([1]))

    """

    def __init__(self, formula, tau=None, theta=None) -> None:
        # NOTE: Pitfall here: setting default value e.g. tau=bp.nodes.Gamma()
        #       would ruin everything because of mutability

        self.formula = formula
        """Model formula"""

        self.tau = tau if tau is not None else bp.nodes.Gamma(1e-3, 1e-3)
        """Node for additive noise precision"""

        self.theta = (
            theta if theta is not None else create_gaussian_theta(formula)
        )
        """Node for model parameters"""

    def __len__(self) -> int:
        """Number of model parameters

        """
        return len(sum(self.formula.terms, []))

    @property
    def theta_marginals(self) -> List[bp.nodes.Gaussian]:
        """Nodes for the basis specific marginal distributions

        """
        # TODO: Test that the marginal distributions are correct
        u = self.theta.get_moments()
        mus = utils.unflatten(u[0], self.formula.terms)
        covs = utils.extract_diag_blocks(
            utils.solve_covariance(u),
            self.formula.terms
        )
        return [
            bp.nodes.Gaussian(mu=mu, Lambda=bp.utils.linalg.inv(cov))
            for (mu, cov) in zip(mus, covs)
        ]

    @property
    def mean_theta(self) -> List[np.ndarray]:
        """Mean estimate of model parameters

        Posterior if model is fitted, otherwise prior.

        """
        return utils.listmap(np.array)(
            utils.unflatten(
                self.theta.get_moments()[0],
                self.formula.terms
            )
        )

    @property
    def covariance_theta(self) -> np.ndarray:
        """Covariance estimate of model parameters

        """
        return utils.solve_covariance(self.theta.get_moments())

    @property
    def inv_mean_tau(self) -> np.ndarray:
        """Additive observation noise variance estimate

        """
        return 1 / self.tau.get_moments()[0]

    def theta_marginal(self, i: int) -> bp.nodes.Gaussian:
        """Extract marginal distribution for a specific term

        """
        u = self.theta.get_moments()
        mus = utils.unflatten(u[0], self.formula.terms)
        covs = utils.extract_diag_blocks(
            utils.solve_covariance(u),
            self.formula.terms
        )
        return bp.nodes.Gaussian(
            mu=mus[i],
            Lambda=bp.utils.linalg.inv(covs[i])
        )

    def fit(self, input_data, y, repeat=1000, verbose=False, **kwargs) -> "GAM":
        """Update BayesPy nodes and construct a GAM predictor

        Parameters
        ----------
        input_data : np.ndarray
            Input data
        y : np.ndarray
            Observations
        repeat : int
            BayesPy allowed repetitions in variational Bayes learning
        verbose : bool
            BayesPy logging

        WARNING: Currently mutates the original object's ``theta`` and ``tau``.

        An option to "reset" the original object to prior is to use the method
        ``initialize_from_prior()`` of BayesPy nodes.

        """
        X = self.formula.design_matrix(input_data)
        F = bp.nodes.SumMultiply("i,i", self.theta, X)
        Y = bp.nodes.GaussianARD(F, self.tau)
        Y.observe(y)
        Q = bp.inference.VB(Y, self.theta, self.tau)
        Q.update(repeat=repeat, verbose=verbose, **kwargs)
        return self

    def predict(self, input_data) -> np.ndarray:
        """Calculate mean of posterior predictive at inputs

        Parameters
        ----------
        input_data : np.ndarray

        """
        X = self.formula.design_matrix(input_data)
        return np.dot(X, np.hstack(self.mean_theta))

    def predict_variance(self, input_data) -> Tuple[np.ndarray]:
        """Predict mean and variance

        Parameters
        ----------
        input_data : np.ndarray

        """
        X = self.formula.design_matrix(input_data)
        F = bp.nodes.SumMultiply("i,i", self.theta, X)
        u = F.get_moments()
        return (u[0], np.diag(utils.solve_covariance(u)) + self.inv_mean_tau)

    def predict_variance_theta(self, input_data) -> Tuple[np.ndarray]:
        """Predict observations with variance from model parameters

        Parameters
        ----------
        input_data : np.ndarray

        """
        X = self.formula.design_matrix(input_data)
        F = bp.nodes.SumMultiply("i,i", self.theta, X)
        # Ensuring correct moments
        #
        # F = F._ensure_moments(
        #     F, bp.inference.vmp.nodes.gaussian.GaussianMoments, ndim=0
        # )
        #
        # NOTE: See also bp.plot.plot_gaussian how std can be calculated
        u = F.get_moments()
        return (u[0], np.diag(utils.solve_covariance(u)))

    def predict_marginals(self, input_data) -> List[np.ndarray]:
        """Predict all terms separately

        Parameters
        ----------
        input_data : np.ndarray

        """
        Xs = [
            self.formula.design_matrix(input_data, i)
            for i in range(len(self.formula))
        ]
        return [np.dot(X, c) for (X, c) in zip(Xs, self.mean_theta)]

    def predict_variance_marginals(self, input_data) -> List[Tuple[np.ndarray]]:
        """Predict variance (theta) for marginal parameter distributions

        Parameters
        ----------
        input_data : np.ndarray

        NOTE: Analogous to self.predict_variance_theta but for marginal
        distributions. Adding observation noise does not make sense as we don't
        know how it is splitted among the model terms.

        """
        Xs = [
            self.formula.design_matrix(input_data, i)
            for i in range(len(self.formula))
        ]
        Fs = [
            bp.nodes.SumMultiply("i,i", theta, X)
            for (X, theta) in zip(Xs, self.theta_marginals)
        ]
        mus = [np.dot(X, c) for (X, c) in zip(Xs, self.mean_theta)]
        sigmas = [np.diag(utils.solve_covariance(F.get_moments())) for F in Fs]
        return list(zip(mus, sigmas))

    def predict_marginal(self, input_data, i: int) -> np.ndarray:
        """Predict a term separately

        Parameters
        ----------
        input_data : np.ndarray

        """
        X = self.formula.design_matrix(input_data, i)
        return np.dot(X, self.mean_theta[i])

    def predict_variance_marginal(
            self,
            input_data,
            i: int
    ) -> Tuple[np.ndarray]:
        """Predict marginal distributions means and variances

        Parameters
        ----------
        input_data : np.ndarray

        """
        # Not refactored with predict_marginal for perf reasons
        X = self.formula.design_matrix(input_data, i)
        F = bp.nodes.SumMultiply("i,i", self.theta_marginal(i), X)
        mu = np.dot(X, self.mean_theta[i])
        sigma = np.diag(utils.solve_covariance(F.get_moments()))
        return (mu, sigma)

    def marginal_residuals(self, input_data, y) -> List[np.ndarray]:
        """Marginal (partial) residuals

        Parameters
        ----------
        input_data : np.ndarray
            Input data
        y : np.ndarray
            Observations

        """
        mus = self.predict_marginals(input_data)
        return [
            y - np.sum(mus[:i] + mus[i + 1:], axis=0)
            for i in range(len(mus))
        ]

    def marginal_residual(self, input_data, y, i: int) -> np.ndarray:
        """Calculate marginal residual for a given term

        Parameters
        ----------
        input_data : np.ndarray

        """
        # Not refactored with marginal_residuals for perf reasons
        mus = self.predict_marginals(input_data)
        return y - np.sum(mus[:i] + mus[i + 1:], axis=0)

    def save(self, filepath: str) -> None:
        """Save the model to disk

        Supported file formats: JSON and HDF5

        """
        file_ext = filepath.split(".")[-1]
        if file_ext in ("h5", "hdf5"):
            with h5py.File(filepath, "w") as h5f:
                group = h5f.create_group("nodes")
                self.tau._save(group.create_group("tau"))
                self.theta._save(group.create_group("theta"))
        elif file_ext == "json":
            with open(filepath, "w+") as jsonf:
                json.dump(
                    obj={
                        "theta": utils.jsonify(self.theta),
                        "tau": utils.jsonify(self.tau)
                    },
                    fp=jsonf
                )
        else:
            raise ValueError(f"Unknown file type: {file_ext}")

    def load(self, filepath: str, **kwargs) -> "GAM":
        """Load model from a file on disk

        """
        file_ext = filepath.split(".")[-1]
        if file_ext in ("h5", "hdf5"):
            with h5py.File(filepath, "r") as h5f:
                tau = self.tau
                tau._load(h5f["nodes"]["tau"])
                theta = create_gaussian_theta(self.formula)
                theta._load(h5f["nodes"]["theta"])
                return GAM(
                    formula=self.formula,
                    tau=tau,
                    theta=theta
                )
        elif file_ext == "json":
            with open(filepath, "r") as jsonf:
                raw = json.load(jsonf)
                tau = utils.set_from_json(raw["tau"], self.tau)
                theta = utils.set_from_json(raw["theta"], self.theta)
        else:
            raise ValueError("Unknown file type: {0}".format(file_ext))
        return GAM(formula=self.formula, tau=tau, theta=theta)


def LinearModel(n_features, prior=None, **kwargs):
    prior = (
        (np.zeros(n_features), 1e-6 * np.identity(n_features)) if prior is None
        else prior
    )
    basis = [(lambda t: t) for i in range(n_features)]
    formula = Formula(terms=[basis], prior=prior)
    return GAM(formula, **kwargs)



"""Top-level configuration for PyTest

"""

import pytest

import gammy


@pytest.fixture(scope="module", params=[
    lambda formula, *args: lambda input_data, y: gammy.bayespy.GAM(
        # HACK: Let's use *args to allow feeding the `tau` parameter
        # only for `gammy.numpy.GAM`. Thus, we leave *args unused for
        # `gammy.bayespy.GAM`.
        formula
    ).fit(input_data, y),
    lambda formula, *args: lambda input_data, y: gammy.numpy.GAM(
        formula, *args
    ).fit(input_data, y)
])
def fit_model(request):
    return request.param



"""Numpy engine

"""

import json
from typing import List, Tuple

import h5py
import numpy as np

import gammy
from gammy import utils
from gammy.formulae import Formula


class Gaussian:
    """Moment calculator for a Gaussian distribution

    Parameters
    ----------
    mu : np.ndarray
        Mean value
    Lambda : np.ndarray
        Precision matrix (inverse of covariance)

    """

    def __init__(self, mu, Lambda) -> None:

        self.mu = mu
        """Mean vector"""

        self.Lambda = Lambda
        """Precision matrix"""

        return

    def get_moments(self) -> List[np.ndarray]:
        return [self.mu, np.linalg.inv(self.Lambda) + np.outer(self.mu, self.mu)]


class Delta:
    """Moment calculator for a Dirac delta distribution

    Parameters
    ----------
    mu : np.ndarray
        Mean value

    """

    def __init__(self, mu: float):

        self.mu = mu
        """Mean vector"""

        return

    def get_moments(self) -> List[np.ndarray]:
        return [np.array(self.mu), np.array(self.mu) ** 2]


def create_gaussian_theta(formula: Formula):
    return Gaussian(*formula.prior)


class GAM:
    """Generalized additive model with NumPy backend

    Parameters
    ----------
    formula : gammy.formulae.Formula
        Formula object containing the terms and prior
    theta : Gaussian
        Model parameters vector
    tau : Delta
        Observation noise precision (inverse variance)

    """

    def __init__(self, formula, tau, theta=None) -> None:

        self.formula = formula
        """Model formula"""

        self.tau = tau
        """Additive noise precision parameter"""

        self.theta = (
            theta if theta is not None else create_gaussian_theta(formula)
        )
        """Model parameters"""

    def __len__(self) -> int:
        """Number of model parameters

        """
        return len(sum(self.formula.terms, []))

    @property
    def theta_marginals(self) -> List[Gaussian]:
        """Marginal distributions of model parameters

        """
        u = self.theta.get_moments()
        mus = utils.unflatten(u[0], self.formula.terms)
        covs = utils.extract_diag_blocks(
            utils.solve_covariance(u),
            self.formula.terms
        )
        return [
            Gaussian(mu=mu, Lambda=np.linalg.inv(cov))
            for (mu, cov) in zip(mus, covs)
        ]

    @property
    def mean_theta(self) -> List[np.ndarray]:
        """Mean estimate of model parameters

        Posterior if model is fitted, otherwise prior.

        """
        return utils.listmap(np.array)(
            utils.unflatten(
                self.theta.get_moments()[0],
                self.formula.terms
            )
        )

    @property
    def covariance_theta(self) -> np.ndarray:
        """Covariance estimate of model parameters

        """
        return utils.solve_covariance(self.theta.get_moments())

    @property
    def inv_mean_tau(self) -> np.ndarray:
        """Additive observation noise variance estimate

        """
        return 1 / self.tau.get_moments()[0]

    def theta_marginal(self, i: int) -> Gaussian:
        """Extract marginal distribution for a specific term

        """
        u = self.theta.get_moments()
        mus = utils.unflatten(u[0], self.formula.terms)
        covs = utils.extract_diag_blocks(
            utils.solve_covariance(u),
            self.formula.terms
        )
        return Gaussian(
            mu=mus[i],
            Lambda=np.linalg.inv(covs[i])
        )

    def fit(self, input_data, y) -> "GAM":
        """Estimate model parameters

        Parameters
        ----------
        input_data : np.ndarray
            Input data
        y : np.ndarray
            Observations

        """
        X = self.formula.design_matrix(input_data)
        #
        # NOTE: Posterior covariance formula based on Kaipio--Somersalo; Remark
        # after Theorem 3.7
        #
        # Perhaps we should use the other posterior covariance formulation which
        # doesn't require directly inverting any matrix. Then we could get rid
        # of all calls of `np.linalg.inv` as long as we refactor out references
        # to the precision matrices `Lambda`. The downside is we would need to
        # refactor also `gammy.models.bayespy.GAM` if want to keep similar
        # interfaces.
        #
        Lambda_post = self.theta.Lambda + self.tau.mu * np.dot(X.T, X)
        mu_post = np.linalg.solve(
            Lambda_post,
            np.dot(self.tau.mu * X.T, y) + np.dot(self.theta.Lambda, self.theta.mu)
        )
        return GAM(
            formula=self.formula,
            tau=self.tau,
            theta=Gaussian(mu=mu_post, Lambda=Lambda_post)
        )

    def predict(self, input_data) -> np.ndarray:
        """Calculate mean of posterior predictive at inputs

        Parameters
        ----------
        input_data : np.ndarray

        """
        X = self.formula.design_matrix(input_data)
        return np.dot(X, np.hstack(self.mean_theta))

    def predict_variance(self, input_data) -> Tuple[np.ndarray]:
        """Predict mean and variance

        Parameters
        ----------
        input_data : np.ndarray

        """
        X = self.formula.design_matrix(input_data)
        Sigma = utils.solve_covariance(self.theta.get_moments())
        return (
            np.dot(X, self.theta.mu),
            # Based on formula: var(A x) = A var(x) A'
            np.diag(np.dot(X, np.dot(Sigma, X.T))) + self.inv_mean_tau
        )

    def predict_variance_theta(self, input_data) -> Tuple[np.ndarray]:
        """Predict observations with variance from model parameters

        Parameters
        ----------
        input_data : np.ndarray

        """
        X = self.formula.design_matrix(input_data)
        Sigma = utils.solve_covariance(self.theta.get_moments())
        return (
            np.dot(X, self.theta.mu),
            np.diag(np.dot(X, np.dot(Sigma, X.T)))
        )

    def predict_marginals(self, input_data) -> List[np.ndarray]:
        """Predict all terms separately

        Parameters
        ----------
        input_data : np.ndarray

        """
        Xs = [
            self.formula.design_matrix(input_data, i)
            for i in range(len(self.formula))
        ]
        return [np.dot(X, c) for (X, c) in zip(Xs, self.mean_theta)]

    def predict_variance_marginals(self, input_data) -> List[Tuple[np.ndarray]]:
        """Predict variance (theta) for marginal parameter distributions

        Parameters
        ----------
        input_data : np.ndarray

        """
        Xs = [
            self.formula.design_matrix(input_data, i)
            for i in range(len(self.formula))
        ]
        Sigmas = [
            utils.solve_covariance(theta.get_moments())
            for theta in self.theta_marginals
        ]
        mus = [np.dot(X, c) for (X, c) in zip(Xs, self.mean_theta)]
        sigmas = [
            np.diag(np.dot(X, np.dot(Sigma, X.T)))
            for (X, Sigma) in zip(Xs, Sigmas)
        ]
        return list(zip(mus, sigmas))

    def predict_marginal(self, input_data, i: int) -> np.ndarray:
        """Predict a term separately

        Parameters
        ----------
        input_data : np.ndarray

        """
        X = self.formula.design_matrix(input_data, i)
        return np.dot(X, self.mean_theta[i])

    def predict_variance_marginal(
            self,
            input_data,
            i: int
    ) -> Tuple[np.ndarray]:
        """Evaluate mean and variance for a given term

        Parameters
        ----------
        input_data : np.ndarray

        """
        X = self.formula.design_matrix(input_data, i)
        Sigma = utils.solve_covariance(self.theta_marginal(i).get_moments())
        mu = np.dot(X, self.mean_theta[i])
        sigma = np.diag(np.dot(X, np.dot(Sigma, X.T)))
        return (mu, sigma)

    def marginal_residuals(self, input_data, y) -> List[np.ndarray]:
        """Marginal (partial) residuals

        Parameters
        ----------
        input_data : np.ndarray
            Input data
        y : np.ndarray
            Observations

        """
        mus = self.predict_marginals(input_data)
        return [
            y - np.sum(mus[:i] + mus[i + 1:], axis=0)
            for i in range(len(mus))
        ]

    def marginal_residual(self, input_data, y, i: int) -> np.ndarray:
        """Calculate marginal residual for a given term

        Parameters
        ----------
        input_data : np.ndarray
            Input data
        y : np.ndarray
            Observations

        """
        mus = self.predict_marginals(input_data)
        return y - np.sum(mus[:i] + mus[i + 1:], axis=0)

    def save(self, filepath: str) -> None:
        """Save the model to disk

        Supported file formats: JSON and HDF5

        """
        file_ext = filepath.split(".")[-1]
        if file_ext in ("h5", "hdf5"):
            with h5py.File(filepath, "w") as h5f:
                theta_group = h5f.create_group("theta")
                tau_group = h5f.create_group("tau")
                utils.write_to_hdf5(theta_group, self.theta.mu, "mu")
                utils.write_to_hdf5(theta_group, self.theta.Lambda, "Lambda")
                utils.write_to_hdf5(tau_group, self.tau.mu, "mu")
        elif file_ext == "json":
            with open(filepath, "w+") as jsonf:
                json.dump(
                    obj={
                        "theta": {
                            "mu": self.theta.mu.tolist(),
                            "Lambda": self.theta.Lambda.tolist()
                        },
                        "tau": {
                            "mu": self.tau.mu
                        }
                    },
                    fp=jsonf
                )
        else:
            raise ValueError(f"Unknown file type: {file_ext}")
        return

    def load(self, filepath: str) -> "GAM":
        """Load model from a file on disk

        """
        file_ext = filepath.split(".")[-1]
        if file_ext in ("h5", "hdf5"):
            with h5py.File(filepath, "r") as h5f:
                tau = Delta(mu=h5f["tau"]["mu"][...])
                theta = Gaussian(
                    mu=h5f["theta"]["mu"][...],
                    Lambda=h5f["theta"]["Lambda"][...]
                )
        elif file_ext == "json":
            with open(filepath, "r") as jsonf:
                raw = json.load(jsonf)
                tau = Delta(raw["tau"]["mu"])
                theta = Gaussian(
                    mu=raw["theta"]["mu"],
                    Lambda=raw["theta"]["Lambda"]
                )
        else:
            raise ValueError(f"Unknown file type: {file_ext}")
        return GAM(formula=self.formula, theta=theta, tau=tau)


def LinearModel(n_features, prior=None, **kwargs):
    prior = (
        (np.zeros(n_features), 1e-6 * np.identity(n_features)) if prior is None
        else prior
    )
    basis = [(lambda t: t) for i in range(n_features)]
    formula = Formula(terms=[basis], prior=prior)
    return GAM(formula, **kwargs)



"""Importable modules in the package

.. autosummary::
   :toctree: _autosummary

   arraymapper
   formulae
   models
   plot
   utils

"""

from .arraymapper import ArrayMapper, x
from .formulae import *
from .models import *
from .plot import *
from .utils import *



"""Miscellaneous utilities

"""

import functools
from typing import (Callable, Dict, Generator, Iterable, List, Tuple)

import numpy as np
import scipy as sp
from scipy import interpolate


#
# Functional
# ~~~~~~~~~~
#


# TODO: A proper curry
def curryish(f: Callable) -> Callable:
    """Lifted partial application

    """

    def g(*args, **kwargs):
        return functools.partial(f, *args, **kwargs)

    return g


def compose2(f: Callable, g: Callable) -> Callable:
    """Compose two functions

    """

    def h(*args, **kwargs):
        return f(g(*args, **kwargs))

    return h


def lift(func: Callable) -> Callable:
    """Transforms a function into an operator

    ``lift :: (a -> b) -> ((c -> a) -> (c -> b))``

    NOTE: Could add func's *args and **kwargs as arguments

    """
    return lambda f: compose2(func, f)


def rlift(func: Callable) -> Callable:
    """Lift from right

    """
    return lambda f: compose2(f, func)


def compose(*funcs: Callable) -> Callable:
    """Function composition

    """
    return functools.partial(functools.reduce, compose2)(funcs)


def pipe(arg, *funcs: Callable) -> Callable:
    """Piping an object through functions

    """
    return compose(*funcs[::-1])(arg)


listmap = curryish(compose(list, map))
listmap.__doc__ = """Map for lists with partial evaluation

"""

tuplemap = curryish(compose(tuple, map))
tuplemap.__doc__ = """Map for tuples with partial evaluation

"""

listfilter = curryish(compose(list, filter))
listfilter.__doc__ = """Filter for lists with partial evaluation

"""

tuplefilter = curryish(compose(tuple, filter))
tuplefilter.__doc__ = """Filter for tuples with partial evaluation

"""


#
# Iterables
# ~~~~~~~~~
#


def unflatten(x: list, y: list) -> list:
    """Unflatten according to a reference

    Example
    -------

    .. code-block:: python

        unflatten([1, 2, 3], [["a", "b"], ["c"]])
        # [[1, 2], [3]]

    """
    def func(cum, this):
        x_crop, res = cum
        return [
            x_crop[len(this):], res + [x_crop[:len(this)]]
        ]

    return functools.reduce(func, list(y), [list(x), []])[-1]


def extract_diag_blocks(x: np.ndarray, y: list) -> List[np.ndarray]:
    """Extract diagonal blocks from a matrix according to a reference

    """

    def func(cum, this):
        x_crop, res = cum
        return [
            x_crop[len(this):, len(this):],
            res + [x_crop[:len(this), :len(this)]]
        ]

    return functools.reduce(func, list(y), [x, []])[-1]


def extend_spline_grid(grid, order: int) -> np.ndarray:
    """Grid extension for higher order splines

    """
    if order < 1:
        raise ValueError(
            "Spline order = n + 1 where n >= 0 is the polynomial degree"
        )
    return grid if order == 1 else pipe(
        grid,
        lambda x: np.append(
            x, np.diff(x)[-(order - 1):][::-1].cumsum() + x[-1]
        ),
        lambda x: np.append(
            x[0] - np.diff(x)[:(order - 1)].cumsum()[::-1], x
        )
    )


def gen_spline_args_from_grid_ext(grid_ext: np.ndarray, order: int, extrapolate: bool) -> Generator:
    """Spline arguments generator from extended grid

    Parameters
    ----------
    grid_ext : np.ndarray
        Extended grid
    order : int
        Order of the splines
    extrapolate : bool
        Allow smooth(ish) extrapolation

    """
    n = len(grid_ext) - order  # Number of basis functions
    (i_left, i_right) = (
        (1, n - 1) if order == 1 else (order - 1, n - order + 1)
    )
    return (
        (grid_ext[i:i + order + 1],) + (
            (extrapolate, -1) if i < i_left
            else (
                (extrapolate, 1) if i >= i_right else (False, 0)
            )
        )
        for i in range(n)
    )


#
# Basis function generation tools
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# TODO / FIXME: The linear-algebraic stuff below remains unfortunately
# largely untested.
#


def squared_dist(x1, x2) -> np.ndarray:
    """Squared distance matrix for column array of N-dimensional points

    Parameters
    ----------
    x1 : np.ndarray
        1-D Column array
    x2 : np.ndarray
        1-D Column array

    Example
    -------

    .. code-block:: python

        x = np.array([[0], [1], [2]])
        squared_dist(x, x)
        # array([[0, 1, 4],
        #        [1, 0, 1],
        #        [4, 1, 0]])

    """
    return (
        np.sum(x1 ** 2, 1).reshape(-1, 1) +
        np.sum(x2 ** 2, 1) -
        2 * np.dot(x1, x2.T)
    )


def exp_squared(x1, x2, corrlen=1.0, sigma=1.0) -> np.ndarray:
    """Exponential squared kernel function

    """
    return sigma * np.exp(-0.5 / corrlen ** 2 * squared_dist(x1, x2))


def exp_sine_squared(x1, x2, corrlen=1.0, sigma=1.0, period=1.0) -> np.ndarray:
    """Exponential sine squared kernel function

    """
    return sigma * np.exp(
        -2.0 / corrlen ** 2 * np.sin(
            np.pi * np.sqrt(squared_dist(x1, x2)) / period
        ) ** 2
    )


def rational_quadratic(x1, x2, corrlen=1.0, sigma=1.0, alpha=1.0) -> np.ndarray:
    """Rational quadratic kernel function

    """
    return sigma * (
        1 + squared_dist(x1, x2) / 2.0 / alpha / corrlen ** 2
    ) ** -alpha


def ornstein_uhlenbeck(x1, x2, corrlen=1.0, sigma=1.0) -> np.ndarray:
    """Ornstein-Uhlenbeck kernel function

    """
    return sigma * np.exp(
        -np.sqrt(squared_dist(x1, x2)) / corrlen
    )


def white_noise(n_dims: int, sigma=1.0, **unused) -> np.ndarray:
    """White noise kernel function

    """
    return sigma * np.identity(n_dims)


def decompose_covariance(H, energy: float=1.01) -> np.ndarray:
    """Most important eigenvectors of a symmetric positive-definite square matrix

    Parameters
    ----------
    H : np.ndarray
        Symmetric positive-definite square matrix
    energy : float
        Truncate to eigenvalues that sum up to this proportion of the total
        eigenvalue sum. If absolutelu all eigenvectors are needed, give value
        slightly larger than one.

    Ordered with respect of the descending eigenvalues. Each
    eigenvector scaled with ``sqrt(λ)``. For theoretical justification,
    see the section on Gaussian Processes in the package documentation.

    NOTE: In the implementation we use np.linalg.svd instead of np.linalg.eigh
    because the latter sometimes returns slightly negative eigenvalues for
    numerical reasons. In those cases the energy trick doesn't give all
    eigenvectors even if we wanted

    REVIEW: There might be problem with serialization. If there are duplicate
    eigenvalues, then on different machines, the vectors might appear in
    different order.

    """

    #
    # Comparison of np.linalg.eigh and np.linalg.svd
    #
    # (W, V) = np.linalg.eigh(H)
    # (U, S, Vh) = np.linalg.svd(H)
    #
    # Holds up to numerical sanity: V[:, ::-1] == U == Vh.T
    #

    (U, S, Vh) = np.linalg.svd(H)
    crop = (S.cumsum() / S.sum()) <= energy
    return np.dot(U[:, crop], np.sqrt(np.diag(S[crop])))


def interp_arrays1d(v, grid, **kwargs) -> List:
    """Create list of interpolators from a given array

    Parameters
    ----------
    v : np.array
        Each column is a "basis" vector
    grid : np.ndarray
        Interpolation grid

    """
    return [
        interpolate.interp1d(grid, v[:, i], **kwargs)
        for i in range(v.shape[1])
    ]


#
# Files and I/O
# ~~~~~~~~~~~~~
#


def write_to_hdf5(group, data, name):
    """Add data to HDF5 handler

    """
    try:
        group.create_dataset(name, data=data, compression="gzip")
    except TypeError:
        group.create_dataset(name, data=data)
    except ValueError:
        raise ValueError(f"Could not write {data}")


#
# BayesPy related
# ~~~~~~~~~~~~~~~
#


def concat_gaussians(gaussians):
    """Concatenate means and covariances to one Gaussian

    Parameters
    ----------
    gaussians : List[Tuple[np.ndarray]]
        List of mean-precision tuples of each Gaussian

    """
    return (
        np.hstack([g[0] for g in gaussians]),
        sp.linalg.block_diag(*[g[1] for g in gaussians])
    )


def solve_covariance(u) -> np.ndarray:
    """Solve covariance matrix from moments

    Parameters
    ----------
    u : List[np.ndarray]
        List of moments as defined by the ``get_moments()`` method call
        of a BayesPy node object.

    """
    cov = u[1] - np.outer(u[0], u[0])
    return cov if cov.shape != (1, 1) else np.array(cov.sum())


solve_precision = compose(np.linalg.inv, solve_covariance)
solve_precision.__doc__ = """Solve precision matrix from moments

"""


def jsonify(node) -> Dict:
    """Turn a expfamily node into a JSON serializable dict

    """
    return {
        **{
            "u{0}".format(i):
                ui.tolist() for (i, ui) in enumerate(node.u)
        },
        **{
            "observed": node.observed
        },
        **{
            "phi{0}".format(i):
                phii.tolist() for (i, phii) in enumerate(node.phi)
        },
        **{
            "f": node.f.tolist(),
            "g": node.g.tolist()
        }
    }


def set_from_json(raw: dict, node):
    """Set BayesPy node attributes from JSON

    """
    node.u = [
        np.array(raw["u{0}".format(i)]) for i in range(len(node.u))
    ]
    node.observed = raw["observed"]
    node.phi = [
        np.array(raw["phi{0}".format(i)]) for i in range(len(node.phi))
    ]
    node.f = np.array(raw["f"])
    node.g = np.array(raw["g"])
    return node


def peaks(x, y):
    """The MATLAB function

    """
    return (
        3 * (1 - x) ** 2 * np.exp(-(x ** 2) - (y + 1) ** 2) -
        10 * (x / 5 - x ** 3 - y ** 5) * np.exp(-x ** 2 - y ** 2) -
        1 / 3 * np.exp(-(x + 1) ** 2 - y ** 2)
    )



import os
from setuptools import setup, find_packages


with open(os.path.join(os.path.dirname(__file__), "README.md")) as fh:
    long_description = fh.read()


setup(
    name="gammy",
    version="0.4.3",
    author="Stratos Staboulis",
    description="Generalized additive models with a Bayesian twist",
    url="https://github.com/malmgrek/gammy",
    packages=find_packages(exclude=["contrib", "doc", "tests"]),
    install_requires=[
        "numpy>=1.10.0",
        "scipy>=0.13.0",
        "bayespy",
        "h5py",
    ],
    extras_require={
        "test": ["pytest"],
    },
    keywords="bayesian statistics modeling gaussian process splines",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        # MIT License
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
)
